# coding: utf-8
import itertools

import pandas as pd


def dimension_levels(dimension_key, dimension):
    if 'display_field' in dimension:
        return [dimension_key, dimension['display_field']]
    return [dimension_key]


def wrap_list(value):
    return value if isinstance(value, (tuple, list)) else [value]


def slice_first(item):
    if isinstance(item, (tuple, list)):
        return item[0]
    return item


def correct_dimension_level_order(dataframe, display_schema):
    if isinstance(dataframe.index, pd.MultiIndex):
        dimension_orders = [order
                            for key, dimension in display_schema['dimensions'].items()
                            for order in dimension_levels(key, dimension)]

        dataframe = dataframe.reorder_levels(dataframe.index.names.index(level)
                                             for level in dimension_orders)

    metrics = list(display_schema['metrics'])
    if display_schema.get('references'):
        references = [''] + list(display_schema['references'])
        return dataframe[list(itertools.product(references, metrics))]

    return dataframe[metrics]


def filter_duplicates(iterable):
    filtered_list, seen = [], set()
    for item in iterable:
        key = slice_first(item)

        if key in seen:
            continue

        seen.add(key)
        filtered_list.append(item)

    return filtered_list



# coding: utf-8
from pypika import terms
from . import Database


class Round(terms.Function):
    # Wrapper for Vertica ROUND function for rounding dates.

    def __init__(self, field, date_format, alias=None):
        super(Round, self).__init__('ROUND', field, date_format, alias=alias)


class Vertica(Database):
    # Vertica client that uses the vertica_python driver.

    def __init__(self, host='localhost', port=5433, database='vertica',
                 user='vertica', password=None,
                 read_timeout=None):
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password
        self.read_timeout = read_timeout

    def connect(self):
        import vertica_python

        return vertica_python.connect(
            host=self.host, port=self.port, database=self.database,
            user=self.user, password=self.password,
            read_timeout=self.read_timeout,
        )

    def round_date(self, field, interval):
        return Round(field, interval)



# coding: utf-8
import copy
import logging

import pandas as pd

from pypika import Query, Interval, JoinType, functions as fn

logger = logging.Logger('fireant')

reference_dimension_mappers = {
    'yoy': lambda join_key: join_key + Interval(weeks=52),
    'qoq': lambda join_key: join_key + Interval(quarters=1),
    'mom': lambda join_key: join_key + Interval(weeks=4),
    'wow': lambda join_key: join_key + Interval(weeks=1),
}
reference_metric_mappers = {
    'd': lambda field, join_field: (field - join_field),
    'p': lambda field, join_field: ((field - join_field) / fn.NullIf(join_field, 0)),
}


class QueryManager(object):
    def query_data(self, database, table, joins=None,
                   metrics=None, dimensions=None,
                   mfilters=None, dfilters=None,
                   references=None, rollup=None):
        """
        Loads a pandas data frame given a table and a description of the request.

        :param database:
            The database interface to use to execute the connection

       :param table:
            Type: pypika.Table
            (Required) The primary table to select data from.  In SQL, this is the table in the FROM clause.

        :param joins:
            Type: list[tuple[pypika.Table, pypika.Criterion, pypika.JoinType]]
            (Optional) A list of tables to join in the query.  If a metric should be selected from a table besides the
            table parameter, the additional table *must* be joined.

            The tuple includes in the following order: Join table, Join criterion, Join type

        :param metrics:
            Type: dict[str: pypika.Field]]
            A dict containing metrics fields to query.  This value is required to contain at least one entry.  The key used
            will match the corresponding column in the returned pd.DataFrame.

        :param dimensions:
            Type: OrderedDict[str: pypika.Field]]
            (Optional) An OrderedDict containing indices to query.  If empty, a pd.Series will be returned containing one
            value per metric.  If given one entry, a pd.DataFrame with a single index will be returned.  If more than one
            value is given, then a pd.DataFrame with a multi-index will be returned.  The key used will match the
            corresponding index level in the returned DataFrame.

            If a dict is used instead of an OrderedDict, then the index levels cannot be guarenteed, so in most cases an
            OrderedDict should be used.

        :param mfilters:
            Type: list[pypika.Criterion]
            (Optional) A list containing criterions to use in the HAVING clause of the query.  This will reduce the data
            to instances where metrics meet the criteria.

        :param dfilters:
            Type: list[pypika.Criterion]
            (Optional) A list containing criterions to use in the WHERE clause of the query.  This will reduce the data
            to instances where dimensions meet the criteria.

        :param references:
            Type: dict[str: str]]
            (Optional) A dict containing comparison operations.  The key must match a valid reference.  The value must
            match the key of a provided dimension.  The value must match a dimensions of the appropriate type, usually a
            Time Series dimension.

            For each value passed, the query will be duplicated containing values for a comparison.

            References can also modify the comparison field to show values such as change.  The format is as follows:

            {key}_{modifier}

            The modifier type is optional and the underscore must be omitted when one is not used.

            Valid references are as follows:

            - *wow*: Week over Week - duplicates the query for values one week prior.
            - *mom*: Month over Month - duplicates the query for values 4 weeks prior.
            - *qoq*: Quarter over Quarter - duplicates the query for values one quarter prior.
            - *yoy*: Year over Year - duplicates the query for values 52 weeks prior.

            Valid modifier types are as follows:

            - *d*: Delta - difference between previous value and current
            - *p*: Delta Percentage - difference between previous value and current as a percentage of the previous value.

            Examples:

            "wow":   Week over Week
            "wow_d": delta Week over Week

        :param rollup:
            Type: list[str]
            (Optional) Lists the dimensions to include for ROLLUP which calculates the totals across groups.  This list
            must be a subset of the keys of the dimensions parameter dict.

            .. note::

                When using rollup for less than all of the dimensions, the dimensions included in the ROLLUP will be
                moved after the non-ROLLUP dimensions.

        :return:
            A pd.DataFrame indexed by the provided dimensions paramaters containing columns for each metrics parameter.
        """
        query = self._build_data_query(table, joins or dict(), metrics or dict(), dimensions or dict(),
                                       dfilters or dict(), mfilters or dict(), references or dict(), rollup or dict())

        querystring = str(query)
        logger.info("Executing query:\n----START----\n{query}\n-----END-----".format(query=querystring))

        dataframe = database.fetch_dataframe(querystring)
        dataframe.columns = [col.decode('utf-8') if isinstance(col, bytes) else col
                             for col in dataframe.columns]

        if dimensions:
            dataframe = dataframe.set_index(
                # Removed the reference keys for now
                list(dimensions.keys())  # + ['{1}_{0}'.format(*ref) for ref in references.items()]
            ).sort_index()

        if references:
            dataframe.columns = pd.MultiIndex.from_product([[''] + list(references.keys()), list(metrics.keys())])

        return dataframe

    def query_dimension_options(self, database, table, joins=None, dimensions=None, filters=None, limit=None):
        """
        Builds and executes a query to retrieve possible dimension options given a set of filters.

        :param database:
            The database interface to use to execute the connection
        :param table:
            See above

        :param joins:
            See above

        :param dimensions:
            See above

        :param filters:
            See above

        :param limit:
            An optional limit to the number of results returned.

        :return:
        """
        query = self._build_dimension_query(table, joins, dimensions, filters, limit)
        results = database.fetch(str(query))
        return [{k: v for k, v in zip(dimensions.keys(), result)}
                for result in results]

    def _build_data_query(self, table, joins, metrics, dimensions, dfilters, mfilters, references, rollup):
        args = (table, joins, metrics, dimensions, dfilters, mfilters, rollup)
        query = self._build_query_inner(*args)

        if references:
            return self._build_reference_query(query, references, *args)

        return self._add_sorting(query, list(dimensions.values()))

    def _build_reference_query(self, query, references, table, joins, metrics, dimensions, dfilters, mfilters, rollup):
        wrapper_query = Query.from_(query).select(*[query.field(key).as_(key)
                                                    for key in list(dimensions.keys()) + list(metrics.keys())])

        for reference_key, dimension_key in references.items():
            dimension_f, metric_f = self._get_reference_mappers(reference_key)

            dimensions, dfilters = self._replace_dim_for_ref(dfilters, dimension_key, dimensions,
                                                             dimension_f)
            ref_query = self._build_query_inner(table, joins, metrics, dimensions,
                                                dfilters, mfilters, rollup)

            ref_criteria = query.field(dimension_key) == ref_query.field(dimension_key)
            for dkey in set(dimensions.keys()) - {dimension_key}:
                ref_criteria &= query.field(dkey) == ref_query.field(dkey)

            # Join the reference query and select the reference dimension and all metrics
            # This ignores additional dimensions since they are identical to the primary results
            wrapper_query = (
                wrapper_query.join(ref_query, JoinType.left)
                    .on(ref_criteria)
                    # Don't select this for now
                    # .select(ref_dimension.as_(self._suffix(dimension_key, reference_key)))
                    .select(
                    *[metric_f(query.field(key), ref_query.field(key)).as_(self._suffix(key, reference_key))
                      for key in metrics.keys()])
            )

        return self._add_sorting(wrapper_query, [query.field(dkey) for dkey in dimensions.keys()])

    def _build_dimension_query(self, table, joins, dimensions, filters, limit=None):
        query = Query.from_(table).distinct()
        query = self._add_joins(joins, query)
        query = self._add_filters(query, filters, [])

        for key, dimension in dimensions.items():
            query = query.select(dimension.as_(key))

        if limit:
            query = query[:limit]

        return query

    def _build_query_inner(self, table, joins, metrics, dimensions, dfilters, mfilters, rollup):
        query = Query.from_(table)
        query = self._add_joins(joins, query)
        query = self._select_dimensions(query, dimensions, rollup)
        query = self._select_metrics(query, metrics)
        return self._add_filters(query, dfilters, mfilters)

    @staticmethod
    def _add_joins(joins, query):
        for join_table, criterion, join_type in joins:
            query = query.join(join_table, how=join_type).on(criterion)
        return query

    @staticmethod
    def _select_dimensions(query, dimensions, rollup):
        dims = [dimension.as_(key)
                for key, dimension in dimensions.items()
                if key not in rollup]
        if dims:
            query = query.select(*dims).groupby(*dims)

        rollup_dims = [dimension.as_(key)
                       for key, dimension in dimensions.items()
                       if key in rollup]
        if rollup_dims:
            query = query.select(*rollup_dims).rollup(*rollup_dims)

        return query

    @staticmethod
    def _select_metrics(query, metrics):
        return query.select(*[metric.as_(key)
                              for key, metric in metrics.items()])

    @staticmethod
    def _add_filters(query, dfilters, mfilters):
        if dfilters:
            for dfx in dfilters:
                query = query.where(dfx)

        if mfilters:
            for mfx in mfilters:
                query = query.having(mfx)

        return query

    @staticmethod
    def _add_sorting(query, dimensions):
        if dimensions:
            return query.orderby(*dimensions)
        return query

    @staticmethod
    def _suffix(key, suffix):
        return '%s_%s' % (key, suffix) if suffix else key

    @staticmethod
    def _get_reference_mappers(reference_key):
        """
        Selects the mapper functions for a reference operation.

        :param reference_key:
            A string identifier that maps to the reference operations.  This contains one to two parts separated by an
            underscore '_' character.  The first part is required and maps to a reference criteria from the dict
            reference_criterion.  The second part is optional and maps to a reference field from the dict
            reference_field.

            The reference criteria is the criteria used to link the values from the original query to the comparison
            values.  For example, in wow (Week over Week), the criteria maps a date field to the same date field minus one
            week.

            The reference field is an optional part that modifies the selected field in the final results.  For example,
            in wow_p (Week over Week change as a Percentage), the final field is the query field minus the comparison
            field as a ratio of the join field.  This gives the change in value over the last week as a percentage of
            the current value.  If the reference key does not contain two parts (no '_' character), then a function will
            still be returned for metric_mapper, however it will not modify the final field.

            Concretely, if the field is clicks and current value is 105 and the wow value is 100, then the change is
            (105 - 100) = 5 clicks, then the wow-p value is (105 - 100) / 100 = 0.05.
        :return:
            Type: Tuple
            (dimension_mapper, metric_mapper)
        """
        split_ref = reference_key.split('_')
        if 1 < len(split_ref):
            reference_key, opt_parts = split_ref[0], split_ref[1]
        else:
            reference_key, opt_parts = split_ref[0], None

        return (
            reference_dimension_mappers.get(reference_key, lambda join_key: join_key),
            reference_metric_mappers.get(opt_parts, lambda field, join_field: join_field)
        )

    @staticmethod
    def _build_reference_join_criterion(query, dimensions, criterion_f, dimension_key):
        """
        Returns join criterion for joining a reference query.

        :param query:

        :param dimensions:

        :param criterion_f:

        :param dimension_key:

        :return:
            join criterion for joining a reference query.
        """
        compared_dimension = dimensions[dimension_key]
        cx = criterion_f(compared_dimension, query.field(dimension_key))
        for dimension in dimensions.values():
            if dimension is not compared_dimension:
                cx &= dimension == dimension.for_(query)
        return cx

    @staticmethod
    def _replace_dim_for_ref(dfilters, dimension_key, dimensions, dimension_f):
        """
        Replaces the dimension used by a reference in the dimension schema and dimension filter schema.

        We do this in order to build the same query with a shifted date instead of the actual date.
        """
        target_dimension = dimensions[dimension_key]

        new_dimensions = copy.deepcopy(dimensions)
        new_dimensions[dimension_key] = dimension_f(target_dimension)

        new_dfilters = []
        for dfilter in dfilters:
            # This is a very hack way of doing this.  This expects the dfilter to use the exact same reference to
            # the dimension's definition, which the slicer manager guarentees and the unit tests simulate.
            # TODO provide a utility in pypika for checking if these are the same
            try:
                if dfilter.term.fields()[0] is target_dimension.fields()[0]:
                    dfilter = copy.deepcopy(dfilter)
                    dfilter.term = dimension_f(dfilter.term)
            except (AttributeError, IndexError):
                pass  # If the above if-expression cannot be evaluated, then its not the filter we are looking for

            new_dfilters.append(dfilter)

        return new_dimensions, new_dfilters



# coding: utf-8

from .base import Transformer, TransformationException
from .datatables import DataTablesRowIndexTransformer, DataTablesColumnIndexTransformer
from .datatables import (DataTablesRowIndexTransformer, DataTablesColumnIndexTransformer, CSVRowIndexTransformer,
                         CSVColumnIndexTransformer)
from .highcharts import HighchartsLineTransformer, HighchartsColumnTransformer, HighchartsBarTransformer
from .notebooks import (PandasRowIndexTransformer, PandasColumnIndexTransformer, MatplotlibLineChartTransformer,
                        MatplotlibBarChartTransformer)

notebooks_tx = {
    'row_index_table': PandasRowIndexTransformer(),
    'column_index_table': PandasColumnIndexTransformer(),
    'line_chart': MatplotlibLineChartTransformer(),
    'bar_chart': MatplotlibBarChartTransformer(),
}

bundles = {
    'notebooks': notebooks_tx,
    'highcharts': {
        'line_chart': HighchartsLineTransformer(),
        'column_chart': HighchartsColumnTransformer(),
        'bar_chart': HighchartsBarTransformer(),
    },
    'datatables': {
        'row_index_table': DataTablesRowIndexTransformer(),
        'column_index_table': DataTablesColumnIndexTransformer(),
        'row_index_csv': CSVRowIndexTransformer(),
        'column_index_csv': CSVColumnIndexTransformer(),
    },
}



# coding: utf-8
import logging

import pandas as pd

logger = logging.getLogger(__name__)


class Database(object):
    def connect(self):
        raise NotImplementedError

    def round_date(self, field, interval):
        raise NotImplementedError

    def fetch(self, query):
        with self.connect() as connection:
            cursor = connection.cursor()
            cursor.execute(query)
            return cursor.fetchall()

    def fetch_dataframe(self, query):
        with self.connect() as connection:
            return pd.read_sql(query, connection)



# coding: utf-8

class Reference(object):
    def __init__(self, element_key):
        self.element_key = element_key


class WoW(Reference):
    key = 'wow'
    label = 'WoW'


class MoM(Reference):
    key = 'mom'
    label = 'MoM'


class QoQ(Reference):
    key = 'qoq'
    label = 'QoQ'


class YoY(Reference):
    key = 'yoy'
    label = 'YoY'


class Delta(Reference):
    def __init__(self, reference):
        super(Delta, self).__init__(reference.element_key)
        self.key = '%s_d' % reference.key
        self.label = '%s Δ' % reference.label


class DeltaPercentage(Reference):
    def __init__(self, reference):
        super(DeltaPercentage, self).__init__(reference.element_key)
        self.key = '%s_p' % reference.key
        self.label = '%s Δ%%' % reference.label



# coding: utf-8
import pandas as pd

from fireant import utils


class WidgetGroupManager(object):
    def __init__(self, widget_group):
        self.widget_group = widget_group

    def render(self, dimensions=None, metric_filters=None, dimension_filters=None, references=None, operations=None):
        combined_dimensions = utils.filter_duplicates(self.widget_group.dimensions + (dimensions or []))
        combined_references = utils.filter_duplicates(self.widget_group.references + (references or []))
        combined_operations = utils.filter_duplicates(self.widget_group.operations + (operations or []))

        dataframe = self.widget_group.slicer.manager.data(
            metrics=[metric
                     for widget in self.widget_group.widgets
                     for metric in widget.metrics],
            dimensions=combined_dimensions,
            metric_filters=metric_filters or [],
            dimension_filters=self.widget_group.dimension_filters + (dimension_filters or []),
            references=combined_references,
            operations=combined_operations,
        )

        return list(self._transform_widgets(self.widget_group.widgets, dataframe,
                                            combined_dimensions, combined_references))

    def _transform_widgets(self, widgets, dataframe, dimensions, references):
        for widget in widgets:
            display_schema = self.widget_group.slicer.manager.display_schema(
                metrics=widget.metrics,
                dimensions=dimensions,
                references=references,
            )

            if display_schema['references']:
                widget_columns = [(reference, metric)
                                  for reference in [''] + list(display_schema['references'].keys())
                                  for metric in widget.metrics]
            else:
                widget_columns = widget.metrics

            widget_df = utils.correct_dimension_level_order(pd.DataFrame(dataframe[widget_columns]), display_schema)
            yield widget.transformer.transform(widget_df, display_schema)



# coding: utf-8
import numpy as np
import pandas as pd

from fireant import settings, utils
from .base import Transformer, TransformationException

colors = {
    'kayak': ['#FF690F', '#1A1A1A', '#3083F0', '#00B86B', '#D10244', '#FFDBE5', '#7B4F4B', '#B903AA', '#B05B6F'],
    'dark-blue': ["#DDDF0D", "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                  "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'dark-green': ["#DDDF0D", "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                   "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'dark-unica': ["#2B908F", "#90EE7E", "#F45B5B", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                   "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'gray': ["#DDDF0D", "#7798BF", "#55BF3B", "#DF5353", "#AAEEEE", "#FF0066", "#EEAAEE",
             "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'grid-light': ["#7CB5EC", "#F7A35C", "#90EE7E", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                   "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'grid': ['#058DC7', '#50B432', '#ED561B', '#DDDF00', '#24CBE5', '#64E572', '#FF9655', '#FFF263', '#6AF9C4'],
    'sand-signika': ["#F45B5B", "#8085E9", "#8D4654", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                     "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'skies': ["#514F78", "#42A07B", "#9B5E4A", "#72727F", "#1F949A", "#82914E", "#86777F", "#42A07B"],
}


def _format_data_point(value):
    if isinstance(value, pd.Timestamp):
        return int(value.asm8) // int(1e6)
    if value is None or (isinstance(value, (float, int)) and np.isnan(value)):
        return None
    if isinstance(value, np.int64):
        # Cannot serialize np.int64 to json
        return int(value)
    return value


class HighchartsLineTransformer(Transformer):
    """
    Transforms data frames into Highcharts format for several chart types, particularly line or bar charts.
    """

    chart_type = 'line'

    def prevalidate_request(self, slicer, metrics, dimensions,
                            metric_filters, dimension_filters,
                            references, operations):
        super(HighchartsLineTransformer, self).prevalidate_request(slicer, metrics, dimensions,
                                                                   metric_filters, dimension_filters,
                                                                   references, operations)

        from fireant.slicer import ContinuousDimension

        dimension0 = slicer.dimensions[dimensions[0]]
        if not dimensions or not isinstance(dimension0, ContinuousDimension):
            raise TransformationException('Highcharts line charts require a continuous dimension as the first '
                                          'dimension.  Please add a continuous dimension from your Slicer to '
                                          'your request.')

    def transform(self, dataframe, display_schema):
        has_references = isinstance(dataframe.columns, pd.MultiIndex)

        dim_ordinal = {name: ordinal
                       for ordinal, name in enumerate(dataframe.index.names)}
        dataframe = self._prepare_dataframe(dataframe, dim_ordinal, display_schema['dimensions'])

        if has_references:
            series = sum(
                [self._make_series(dataframe[level], dim_ordinal, display_schema, reference=level or None)
                 for level in dataframe.columns.levels[0]],
                []
            )

        else:
            series = self._make_series(dataframe, dim_ordinal, display_schema)

        result = {
            'chart': {'type': self.chart_type, 'zoomType': 'x'},
            'title': {'text': None},
            'xAxis': self.xaxis_options(dataframe, dim_ordinal, display_schema),
            'yAxis': self.yaxis_options(dataframe, dim_ordinal, display_schema),
            'tooltip': {'shared': True},
            'series': series
        }

        return result

    def xaxis_options(self, dataframe, dim_ordinal, display_schema):
        return {
            'type': 'datetime' if isinstance(dataframe.index, pd.DatetimeIndex) else 'linear'
        }

    def yaxis_options(self, dataframe, dim_ordinal, display_schema):
        return [{
            'title': None
        }] * len(display_schema['metrics'])

    def _make_series(self, dataframe, dim_ordinal, display_schema, reference=None):
        metrics = list(dataframe.columns.levels[0]
                       if isinstance(dataframe.columns, pd.MultiIndex)
                       else dataframe.columns)

        color = colors.get(settings.highcharts_colors, 'grid')
        n_colors = len(color)

        return [self._make_series_item(idx, item, dim_ordinal, display_schema, metrics, reference, color[i % n_colors])
                for i, (idx, item) in enumerate(dataframe.iteritems())]

    def _make_series_item(self, idx, item, dim_ordinal, display_schema, metrics, reference, color='#000'):
        metric_key = utils.slice_first(idx)
        return {
            'name': self._format_label(idx, dim_ordinal, display_schema, reference),
            'data': self._format_data(item),
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'yAxis': metrics.index(utils.slice_first(idx)),
            'color': color,
            'dashStyle': 'Dot' if reference else 'Solid'
        }

    @staticmethod
    def _make_categories(dataframe, dim_ordinal, display_schema):
        return None

    def _format_tooltip(self, metric_schema):
        tooltip = {}

        if 'precision' in metric_schema:
            tooltip['valueDecimals'] = metric_schema['precision']
        if 'prefix' in metric_schema:
            tooltip['valuePrefix'] = metric_schema['prefix']
        if 'suffix' in metric_schema:
            tooltip['valueSuffix'] = metric_schema['suffix']

        return tooltip

    def _prepare_dataframe(self, dataframe, dim_ordinal, dimensions):
        # Replaces invalid values and unstacks the data frame for line charts.

        # Force all fields to be float (Safer for highcharts)
        dataframe = dataframe.astype(np.float).replace([np.inf, -np.inf], np.nan)

        # Unstack multi-indices
        if 1 < len(dimensions):
            # We need to unstack all of the dimensions here after the first dimension, which is the first dimension in
            # the dimensions list, not necessarily the one in the dataframe
            unstack_levels = list(self._unstack_levels(list(dimensions.items())[1:], dim_ordinal))
            dataframe = dataframe.unstack(level=unstack_levels)

        return dataframe

    def _format_label(self, idx, dim_ordinal, display_schema, reference):
        is_multidimensional = isinstance(idx, tuple)
        if is_multidimensional:
            metric = display_schema['metrics'].get(idx[0], idx[0])
        else:
            metric = display_schema['metrics'].get(idx, idx)

        metric_label = metric['label']
        if reference:
            metric_label += ' {reference}'.format(
                reference=display_schema['references'][reference]
            )

        if not is_multidimensional:
            return metric_label

        dimension_labels = [self._format_dimension_display(dim_ordinal, key, dimension, idx)
                            for key, dimension in list(display_schema['dimensions'].items())[1:]]
        dimension_labels = [dimension_label  # filter out the NaNs
                            for dimension_label in dimension_labels
                            if dimension_label is not np.nan]

        return (
            '{metric} ({dimensions})'.format(
                metric=metric_label,
                dimensions=', '.join(map(str, dimension_labels))
            )
            if dimension_labels
            else metric_label
        )

    @staticmethod
    def _format_dimension_display(dim_ordinal, key, dimension, idx):
        if 'display_field' in dimension:
            display_field = dimension['display_field']
            return idx[dim_ordinal[display_field]]

        dimension_value = idx[dim_ordinal[key]]
        if 'display_options' in dimension:
            dimension_value = dimension['display_options'].get(dimension_value, dimension_value)
        return dimension_value

    def _format_data(self, column):
        if isinstance(column, float):
            return [_format_data_point(column)]

        return [self._format_point(key, value)
                for key, value in column.iteritems()
                if not (isinstance(value, (float, int)) and np.isnan(value))]

    @staticmethod
    def _format_point(x, y):
        return (_format_data_point(x), _format_data_point(y))

    def _unstack_levels(self, dimensions, dim_ordinal):
        for key, dimension in dimensions:
            yield dim_ordinal[key]

            if 'display_field' in dimension:
                yield dim_ordinal[dimension['display_field']]


class HighchartsColumnTransformer(HighchartsLineTransformer):
    chart_type = 'column'

    def prevalidate_request(self, slicer, metrics, dimensions,
                            metric_filters, dimension_filters,
                            references, operations):
        super(HighchartsLineTransformer, self).prevalidate_request(slicer, metrics, dimensions,
                                                                   metric_filters, dimension_filters,
                                                                   references, operations)

        if dimensions and 2 < len(dimensions):
            # Too many dimensions
            raise TransformationException('Highcharts bar and column charts support at a maximum two dimensions.  '
                                          'Request included %d dimensions.' % len(dimensions))

        if dimensions and 2 == len(dimensions) and metrics and 1 < len(metrics):
            # Too many metrics
            raise TransformationException('Highcharts bar and column charts support at a maximum one metric with '
                                          'two dimensions.  Please remove some dimensions or metrics.  '
                                          'Request included %d metrics and %d dimensions.' % (len(metrics),
                                                                                              len(dimensions)))

    def _make_series_item(self, idx, item, dim_ordinal, display_schema, metrics, reference, color='#000'):
        metric_key = utils.slice_first(idx)
        return {
            'name': self._format_label(idx, dim_ordinal, display_schema, reference),
            'data': [_format_data_point(x)
                     for x in item
                     if not (isinstance(x, (float, int)) and np.isnan(x))],
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'yAxis': metrics.index(metric_key),
            'color': color,
        }

    def xaxis_options(self, dataframe, dim_ordinal, display_schema):
        result = {'type': 'categorical'}

        categories = self._make_categories(dataframe, dim_ordinal, display_schema)
        if categories is not None:
            result['categories'] = categories

        return result

    def _validate_dimensions(self, dataframe, dimensions):
        if 1 < len(dimensions) and 1 < len(dataframe.columns):
            raise TransformationException('Cannot transform %s chart.  '
                                          'No more than 1 dimension or 2 dimensions '
                                          'with 1 metric are allowed.' % self.chart_type)

    def _prepare_dataframe(self, dataframe, dim_ordinal, dimensions):
        # Replaces invalid values and unstacks the data frame for line charts.

        # Force all fields to be float (Safer for highcharts)
        dataframe = dataframe.replace([np.inf, -np.inf], np.nan)

        # Unstack multi-indices
        if 1 < len(dimensions):
            unstack_levels = list(self._unstack_levels(list(dimensions.items())[1:], dim_ordinal))
            dataframe = dataframe.unstack(level=unstack_levels)

        return dataframe

    @staticmethod
    def _make_categories(dataframe, dim_ordinal, display_schema):
        if not display_schema['dimensions']:
            return None

        category_dimension = list(display_schema['dimensions'].values())[0]
        if 'display_options' in category_dimension:
            return [category_dimension['display_options'].get(dim, dim)
                    # Pandas gives both NaN or None in the index depending on whether a level was unstacked
                    if dim and not (isinstance(dim, (float, int)) and np.isnan(dim))
                    else 'Totals'
                    for dim in dataframe.index]

        if 'display_field' in category_dimension:
            display_field = category_dimension['display_field']
            return dataframe.index.get_level_values(display_field).unique().tolist()


class HighchartsBarTransformer(HighchartsColumnTransformer):
    chart_type = 'bar'
